@@ -67,6 +67,7 @@ static const char *usage =
 "-t, --ftype file-path Prints out the mime type of a file\n"
 "-l, --list            Prints a list of the daemon's rules with numbers\n"
 "-u, --update          Notifies fapolicyd to perform update of database\n"
+"-r, --reload-rules    Notifies fapolicyd to perform reload of rules\n"
 ;
 
 static struct option long_opts[] =
@@ -83,11 +84,15 @@ static struct option long_opts[] =
 	{"ftype",	1, NULL, 't'},
 	{"list",	0, NULL, 'l'},
 	{"update",	0, NULL, 'u'},
+	{"reload-rules",	0, NULL, 'r'},
 	{ NULL,		0, NULL, 0 }
 };
 
 volatile atomic_bool stop = 0;  // Library needs this
-unsigned int debug = 0;			// Library needs this
+unsigned int debug_mode = 0;			// Library needs this
+unsigned int permissive = 0;			// Library needs this
+
+typedef enum _reload_code { DB, RULES} reload_code;
 
 static char *get_line(FILE *f, unsigned *lineno)
 {
@@ -452,7 +457,7 @@ next_iteration:
 }
 
 
-static int do_update(void)
+static int do_reload(int code)
 {
 	int fd = -1;
 	struct stat s;
@@ -487,7 +492,16 @@ static int do_update(void)
 		}
 	}
 
-	ssize_t ret = write(fd, "1\n", 2);
+	ssize_t ret = 0;
+	char str[32] = {0};
+
+	if (code == DB) {
+		snprintf(str, 32, "%c\n", RELOAD_TRUSTDB_COMMAND);
+		ret = write(fd, "1\n", strlen(str));
+	} else if (code == RULES) {
+		snprintf(str, 32, "%c\n", RELOAD_RULES_COMMAND);
+		ret = write(fd, "3\n", strlen(str));
+	}
 
 	if (ret == -1) {
 		fprintf(stderr,"Write: %s -> %s\n", fifo_path, strerror(errno));
@@ -589,10 +603,19 @@ static int check_watch_fs(void)
 		return 1;
 	}
 
+	fd_fgets_context_t *fd_fgets_context = fd_fgets_init();
+	if (!fd_fgets_context) {
+		fprintf(stderr, "Failed fd_fgets_init\n");
+		free_daemon_config(&config);
+		list_empty(&fs);
+		close(fd);
+		return 1;
+	}
+
 	// Build the list of mount point types
 	list_init(&mnt);
 	do {
-		if (fd_fgets(buf, sizeof(buf), fd)) {
+		if (fd_fgets(fd_fgets_context, buf, sizeof(buf), fd)) {
 			sscanf(buf, "%1024s %4096s %31s %127s %d %d\n",
 			       device,point, type, mntops, &fs_req, &fs_passno);
 			// Some file systems are not watchable
@@ -600,9 +623,11 @@ static int check_watch_fs(void)
 				continue;
 			list_append(&mnt, strdup(type), strdup("0"));
 		}
-	} while (!fd_fgets_eof());
+	} while (!fd_fgets_eof(fd_fgets_context));
+	fd_fgets_destroy(fd_fgets_context);
 	close(fd);
 
+
 	// Now search the list we just built
 	for (list_item_t *lptr = mnt.first; lptr; lptr = lptr->next) {
 		// See if the file system is watched
@@ -623,6 +648,7 @@ static int check_watch_fs(void)
 
 	free_daemon_config(&config);
 	list_empty(&fs);
+	list_empty(&mnt);
 	if (found == 0)
 		printf("Nothing appears missing\n");
 
@@ -648,7 +674,7 @@ static int verify_file(const char *path, off_t size, const char *sha)
 	if (sb.st_size != size)
 		warn_size = 1;
 
-	char *sha_buf = get_hash_from_fd2(fd, sb.st_size);
+	char *sha_buf = get_hash_from_fd2(fd, sb.st_size, 1);
 	close(fd);
 
 	if (sha_buf == NULL || strcmp(sha, sha_buf))
@@ -791,13 +817,18 @@ next:
 static int do_status_report(void)
 {
 	const char *reason = "no pid file";
+
+	fd_fgets_context_t *fd_fgets_context = fd_fgets_init();
+	if (!fd_fgets_context)
+		return 1;
+
 	// open pid file
 	int pidfd = open(pidfile, O_RDONLY);
 	if (pidfd >= 0) {
 		char pid_buf[16];
 
 		// read contents
-		if (fd_fgets(pid_buf, sizeof(pid_buf), pidfd)) {
+		if (fd_fgets(fd_fgets_context, pid_buf, sizeof(pid_buf), pidfd)) {
 			int rpt_fd;
 			unsigned int pid, tries = 0;
 			char exe_buf[64];
@@ -829,7 +860,8 @@ static int do_status_report(void)
 
 			// Access a file to provoke a response
 			int fd = open(CONFIG_FILE, O_RDONLY);
-			close(fd);
+			if (fd >= 0)
+				close(fd);
 
 retry:
 			// wait for it
@@ -846,22 +878,25 @@ retry:
 					goto err_out;
 				}
 			}
+			fd_fgets_rewind(fd_fgets_context);
 			do {
 				char buf[80];
-				if (fd_fgets(buf, sizeof(buf), rpt_fd))
+				if (fd_fgets(fd_fgets_context, buf, sizeof(buf), rpt_fd))
 					write(1, buf, strlen(buf));
-			} while (!fd_fgets_eof());
+			} while (!fd_fgets_eof(fd_fgets_context));
 			close(rpt_fd);
 		} else
 			reason = "can't read pid file";
 		close(pidfd);
+		fd_fgets_destroy(fd_fgets_context);
 		return 0;
 	}
 err_out:
-		if (pidfd >= 0)
-			close(pidfd);
-		printf("Can't find fapolicyd: %s\n", reason);
-		return 1;
+	fd_fgets_destroy(fd_fgets_context);
+	if (pidfd >= 0)
+		close(pidfd);
+	printf("Can't find fapolicyd: %s\n", reason);
+	return 1;
 }
 
 int main(int argc, char * const argv[])
@@ -874,7 +909,7 @@ int main(int argc, char * const argv[])
 		return rc;
 	}
 
-	opt = getopt_long(argc, argv, "Ddf:ht:lu",
+	opt = getopt_long(argc, argv, "Ddf:ht:lur",
 				 long_opts, &option_index);
 	switch (opt) {
 	case 'd':
@@ -911,7 +946,12 @@ int main(int argc, char * const argv[])
 	case 'u':
 		if (argc > 2)
 			goto args_err;
-		rc = do_update();
+		rc = do_reload(DB);
+		break;
+	case 'r':
+		if (argc > 2)
+			goto args_err;
+		rc = do_reload(RULES);
 		break;
 
 	// Now the pure long options
