@@ -1,6 +1,6 @@
 /*
  * database.c - Trust database
- * Copyright (c) 2016,2018-22 Red Hat Inc.
+ * Copyright (c) 2016,2018-24 Red Hat Inc.
  * All Rights Reserved.
  *
  * This software may be freely redistributed and/or modified under the
@@ -50,10 +50,11 @@
 #include "backend-manager.h"
 #include "gcc-attributes.h"
 #include "paths.h"
+#include "policy.h"
 
 // Local defines
 enum { READ_DATA, READ_TEST_KEY, READ_DATA_DUP };
-typedef enum { DB_NO_OP, ONE_FILE, RELOAD_DB, FLUSH_CACHE } db_ops_t;
+typedef enum { DB_NO_OP, ONE_FILE, RELOAD_DB, FLUSH_CACHE, RELOAD_RULES } db_ops_t;
 #define BUFFER_SIZE 4096
 #define MEGABYTE	(1024*1024)
 
@@ -71,6 +72,7 @@ static atomic_int reload_db = 0;
 
 static pthread_t update_thread;
 static pthread_mutex_t update_lock;
+static pthread_mutex_t rule_lock;
 
 // Local functions
 static void *update_thread_main(void *arg);
@@ -79,6 +81,7 @@ static int update_database(conf_t *config);
 // External variables
 extern volatile atomic_bool stop;
 extern volatile atomic_bool needs_flush;
+extern volatile atomic_bool reload_rules;
 
 
 static int is_link(const char *path)
@@ -100,6 +103,8 @@ const char *lookup_tsource(unsigned int tsource)
 	{
 	case SRC_RPM:
 		return "rpmdb";
+	case SRC_DEB:
+		return "debdb";
 	case SRC_FILE_DB:
 		return "filedb";
 	}
@@ -111,8 +116,15 @@ int preconstruct_fifo(const conf_t *config)
 	int rc;
 	char err_buff[BUFFER_SIZE];
 
-	/* Make sure that there is no such file/fifo */
-	unlink_fifo();
+	/* Ensure that the RUN_DIR exists */
+	if (mkdir(RUN_DIR, 0770) && errno != EEXIST) {
+		msg(LOG_ERR, "Failed to create a directory %s (%s)", RUN_DIR,
+		    strerror_r(errno, err_buff, BUFFER_SIZE));
+		return 1;
+	} else {
+		/* Make sure that there is no such file/fifo */
+		unlink_fifo();
+	}
 
 	rc = mkfifo(fifo_path, 0660);
 
@@ -186,15 +198,20 @@ static unsigned long pages, max_pages;
 static void close_db(int do_report)
 {
 	if (do_report) {
-		MDB_envinfo stat;
+		MDB_envinfo st;
 
 		// Collect useful stats
 		unsigned size = get_pages_in_use();
-		mdb_env_info(env, &stat);
-		max_pages = stat.me_mapsize / size;
-		msg(LOG_DEBUG, "Trust database max pages: %lu", max_pages);
-		msg(LOG_DEBUG, "Trust database pages in use: %lu (%lu%%)", pages,
-		    max_pages ? ((100*pages)/max_pages) : 0);
+		if (size == 0) {
+			msg(LOG_DEBUG,
+			    "The trust database is empty.");
+		} else {
+			mdb_env_info(env, &st);
+			max_pages = st.me_mapsize / size;
+			msg(LOG_DEBUG, "Trust database max pages: %lu", max_pages);
+			msg(LOG_DEBUG, "Trust database pages in use: %lu (%lu%%)", pages,
+			    max_pages ? ((100*pages)/max_pages) : 0);
+		}
 	}
 
 	// Now close down
@@ -204,12 +221,19 @@ static void close_db(int do_report)
 
 static void check_db_size(void)
 {
-	MDB_envinfo stat;
+	MDB_envinfo st;
 
 	// Collect stats
 	unsigned long size = get_pages_in_use();
-	mdb_env_info(env, &stat);
-	max_pages = stat.me_mapsize / size;
+
+	if (size == 0) {
+		msg(LOG_WARNING,
+		    "The trust database is empty");
+		return;
+	}
+
+	mdb_env_info(env, &st);
+	max_pages = st.me_mapsize / size;
 	unsigned long percent = max_pages ? (100*pages)/max_pages : 0;
 	if (percent > 80)
 		msg(LOG_WARNING, "Trust database at %lu%% capacity - "
@@ -254,7 +278,7 @@ static void abort_transaction(MDB_txn *txn)
  * Convert path to a hash value. Used when the path exceeds the LMDB key
  * limit(511).  Note: Returned value must be deallocated.
  */
-static char *path_to_hash(const char *path, const size_t path_len) MALLOCLIKE;
+static char *path_to_hash(const char *path, const size_t path_len) __attr_dealloc_free __attr_access ((__read_only__, 1, 2));
 static char *path_to_hash(const char *path, const size_t path_len)
 {
 	unsigned char hptr[80];
@@ -381,14 +405,14 @@ static void end_long_term_read_ops(void)
 
 static unsigned get_pages_in_use(void)
 {
-	MDB_stat stat;
+	MDB_stat st;
 
 	start_long_term_read_ops();
-	mdb_stat(lt_txn, dbi, &stat);
+	mdb_stat(lt_txn, dbi, &st);
 	end_long_term_read_ops();
-	pages = stat.ms_leaf_pages + stat.ms_branch_pages +
-		stat.ms_overflow_pages;
-	return stat.ms_psize;
+	pages = st.ms_leaf_pages + st.ms_branch_pages +
+		st.ms_overflow_pages;
+	return st.ms_psize;
 }
 
 // if success, the function returns positive number of entries in database
@@ -410,7 +434,7 @@ static long get_number_of_entries(void)
  * search for the data. It returns NULL on error or if no data found.
  * The returned string must be freed by the caller.
  */
-static char *lt_read_db(const char *index, int operation, int *error) MALLOCLIKE;
+static char *lt_read_db(const char *index, int operation, int *error) __attr_dealloc_free;
 static char *lt_read_db(const char *index, int operation, int *error)
 {
 	int rc;
@@ -483,8 +507,9 @@ static char *lt_read_db(const char *index, int operation, int *error)
 	// A next step might be to check the status field to see that its
 	// trusted.
 	*error = 0;
-	if (operation == READ_TEST_KEY)
-		return (char *)db;
+	if (operation == READ_TEST_KEY) {
+		return strndup(db, MDB_maxkeysize);
+	}
 
 	if ((data = malloc(value.mv_size+1))) {
 		memcpy(data, value.mv_data, value.mv_size);
@@ -835,6 +860,7 @@ int init_database(conf_t *config)
 
 	// update_lock is used in update_database()
 	pthread_mutex_init(&update_lock, NULL);
+	pthread_mutex_init(&rule_lock, NULL);
 
 	if (migrate_database())
 		return 1;
@@ -921,8 +947,14 @@ retry_res:
 
 	res = lt_read_db(path, mode, error);
 
+	// For subjects we do a limited check because the process had to
+	// pass some kind of trust check to even be started and we do not
+	// have an open fd to the file.
 	if (!do_integrity) {
-		return res ? 1 : 0;
+		if (res == NULL)
+			return 0;
+		free(res);
+		return 1;
 	} else {
 		unsigned int tsource;
 		off_t size;
@@ -979,7 +1011,7 @@ retry_res:
 
 			// Calculate a hash only one time
 			if (retry == 1) {
-				hash = get_hash_from_fd2(fd, info->size);
+				hash = get_hash_from_fd2(fd, info->size, 1);
 				if (hash) {
 					strncpy(sha_xattr, hash, 64);
 					sha_xattr[64] = 0;
@@ -1062,6 +1094,7 @@ void close_database(void)
 	// we can close db when we are really sure update_thread does not exist
 	close_db(1);
 	pthread_mutex_destroy(&update_lock);
+	pthread_mutex_destroy(&rule_lock);
 
 	backend_close();
 	unlink_fifo();
@@ -1082,7 +1115,6 @@ void lock_update_thread(void) {
 	//msg(LOG_DEBUG, "lock_update_thread()");
 }
 
-
 /*
  * Unlock wrapper for update mutex
  */
@@ -1091,6 +1123,21 @@ void unlock_update_thread(void) {
 	//msg(LOG_DEBUG, "unlock_update_thread()");
 }
 
+/*
+ * Lock wrapper for rule mutex
+ */
+void lock_rule(void) {
+	pthread_mutex_lock(&rule_lock);
+	//msg(LOG_DEBUG, "lock_rule()");
+}
+
+/*
+ * Unlock wrapper for rule mutex
+ */
+void unlock_rule(void) {
+	pthread_mutex_unlock(&rule_lock);
+	//msg(LOG_DEBUG, "unlock_rule()");
+}
 
 /*
  * This function reloads updated backend db into our internal database.
@@ -1167,7 +1214,7 @@ static int handle_record(const char * buffer)
 	return 0;
 }
 
-void update_trust_database(void)
+void set_reload_trust_database(void)
 {
 	reload_db = 1;
 }
@@ -1210,12 +1257,13 @@ static void *update_thread_main(void *arg)
 	msg(LOG_DEBUG, "Update thread main started");
 #endif
 
-	/* This is a worker thread. Don't handle signals. */
+	/* This is a worker thread. Don't handle external signals. */
 	sigemptyset(&sigs);
 	sigaddset(&sigs, SIGTERM);
 	sigaddset(&sigs, SIGHUP);
+	sigaddset(&sigs, SIGUSR1);
 	sigaddset(&sigs, SIGINT);
-	sigaddset(&sigs, SIGSEGV);
+	sigaddset(&sigs, SIGQUIT);
 	pthread_sigmask(SIG_SETMASK, &sigs, NULL);
 
 	if (ffd[0].fd == 0) {
@@ -1230,6 +1278,14 @@ static void *update_thread_main(void *arg)
 
 		rc = poll(ffd, 1, 1000);
 
+		if (reload_rules) {
+			reload_rules = false;
+			load_rule_file();
+
+			lock_rule();
+			do_reload_rules(config);
+			unlock_rule();
+		}
 		// got SIGHUP
 		if (reload_db) {
 			reload_db = 0;
@@ -1259,9 +1315,10 @@ static void *update_thread_main(void *arg)
 		} else {
 			if (ffd[0].revents & POLLIN) {
 
+				fd_fgets_context_t * fd_fgets_context = fd_fgets_init();
 				do {
-					fd_fgets_rewind();
-					int res = fd_fgets(buff, sizeof(buff), ffd[0].fd);
+					fd_fgets_rewind(fd_fgets_context);
+					int res = fd_fgets(fd_fgets_context, buff, sizeof(buff), ffd[0].fd);
 
 					// nothing to read
 					if (res == -1)
@@ -1286,16 +1343,21 @@ static void *update_thread_main(void *arg)
 								break;
 							}
 
-							if (buff[i] == '1') {
+							if (buff[i] == RELOAD_TRUSTDB_COMMAND) {
 								do_operation = RELOAD_DB;
 								break;
 							}
 
-							if (buff[i] == '2') {
+							if (buff[i] == FLUSH_CACHE_COMMAND) {
 								do_operation = FLUSH_CACHE;
 								break;
 							}
 
+							if (buff[i] == RELOAD_RULES_COMMAND) {
+								do_operation = RELOAD_RULES;
+								break;
+							}
+
 							if (isspace(buff[i]))
 								continue;
 
@@ -1309,6 +1371,14 @@ static void *update_thread_main(void *arg)
 						if (do_operation == RELOAD_DB) {
 							do_operation = DB_NO_OP;
 							do_reload_db(config);
+						} else if (do_operation == RELOAD_RULES) {
+							do_operation = DB_NO_OP;
+
+							load_rule_file();
+
+							lock_rule();
+							do_reload_rules(config);
+							unlock_rule();
 
 							// got "2" -> flush cache
 						} else if (do_operation == FLUSH_CACHE) {
@@ -1321,7 +1391,8 @@ static void *update_thread_main(void *arg)
 						}
 					}
 
-				} while(!fd_fgets_eof());
+				} while(!fd_fgets_eof(fd_fgets_context));
+				fd_fgets_destroy(fd_fgets_context);
 			}
 		}
 	}
@@ -1407,4 +1478,3 @@ void walk_database_finish(void)
 	abort_transaction(lt_txn);
 	close_db(0);
 }
-
