@@ -1,6 +1,6 @@
 /*
  * file.c - functions for accessing attributes of files
- * Copyright (c) 2016,2018-22 Red Hat Inc.
+ * Copyright (c) 2016,2018-23 Red Hat Inc.
  * All Rights Reserved.
  *
  * This software may be freely redistributed and/or modified under the
@@ -32,6 +32,7 @@
 #include <string.h>
 #include <stdlib.h>
 #include <openssl/sha.h>
+#include <openssl/md5.h>
 #include <magic.h>
 #include <libudev.h>
 #include <elf.h>
@@ -52,6 +53,14 @@ magic_t magic_cookie;
 struct cache { dev_t device; const char *devname; };
 static struct cache c = { 0, NULL };
 
+// Local declarations
+static ssize_t safe_read(int fd, char *buf, size_t size)
+				__attr_access ((__write_only__, 2, 3));
+static char *get_program_cwd_from_pid(pid_t pid, size_t blen, char *buf)
+				__attr_access ((__write_only__, 3, 2));
+static void resolve_path(const char *pcwd, char *path, size_t len)
+				__attr_access ((__write_only__, 2, 3));
+
 // readelf -l path-to-app | grep 'Requesting' | cut -d':' -f2 | tr -d ' ]';
 static const char *interpreters[] = {
 	"/lib64/ld-linux-x86-64.so.2",
@@ -220,6 +229,9 @@ char *get_file_from_fd(int fd, pid_t pid, size_t blen, char *buf)
 	char procfd_path[32];
 	ssize_t path_len;
 
+	if (blen == 0)
+		return NULL;
+
 	snprintf(procfd_path, sizeof(procfd_path)-1,
 		"/proc/self/fd/%d", fd);
 	path_len = readlink(procfd_path, buf, blen - 1);
@@ -375,7 +387,6 @@ char *get_file_type_from_fd(int fd, const struct file_info *i, const char *path,
 
 	// Do the normal classification
 	ptr = magic_descriptor(magic_cookie, fd);
-	rewind_fd(fd);
 	if (ptr) {
 		char *str;
 		strncpy(buf, ptr, blen-1);
@@ -397,6 +408,9 @@ char *bytes2hex(char *final, const unsigned char *buf, unsigned int size)
 	char *ptr = final;
 	const char *hex = "0123456789abcdef";
 
+	if (final == NULL)
+		return final;
+
 	for (i=0; i<size; i++) {
 		*ptr++ = hex[(buf[i] & 0xF0)>>4]; /* Upper nibble */
 		*ptr++ = hex[buf[i] & 0x0F];      /* Lower nibble */
@@ -421,29 +435,44 @@ static ssize_t safe_read(int fd, char *buf, size_t size)
 
 /*
  * Given a fd, calculate the hash by accessing size bytes of the file.
+ * Calculate SHA256 by default or compute MD5.
  * Returns a char pointer of the hash which the caller must free.
  * If a size of 0 is passed, it will return a NULL pointer.
  * If there is an error with mmap, it will also return a NULL pointer.
  */
-static const char *degenerate_hash = "e3b0c44298fc1c149afbf4c8996fb92427ae41e4649b934ca495991b7852b855";
-char *get_hash_from_fd2(int fd, size_t size)
+static const char *degenerate_hash_sha = "e3b0c44298fc1c149afbf4c8996fb92427ae41e4649b934ca495991b7852b855";
+static const char *degenerate_hash_md5 = "d41d8cd98f00b204e9800998ecf8427e";
+char *get_hash_from_fd2(int fd, size_t size, const int is_sha)
 {
 	unsigned char *mapped;
 	char *digest = NULL;
 
-	if (size == 0)
-		return strdup(degenerate_hash);
+	if (size == 0) {
+		if (is_sha)
+			return strdup(degenerate_hash_sha);
+		return strdup(degenerate_hash_md5);
+	}
 
 	mapped = mmap(0, size, PROT_READ, MAP_PRIVATE|MAP_POPULATE, fd, 0);
 	if (mapped != MAP_FAILED) {
-		unsigned char hptr[40];
+		const int digest_length = is_sha ? SHA256_DIGEST_LENGTH : 16;
+		// Just use the larger one as buffer.
+		unsigned char hptr[SHA256_DIGEST_LENGTH];
 
-		SHA256(mapped, size, (unsigned char *)&hptr);
+		if (is_sha) {
+			SHA256(mapped, size, (unsigned char *)&hptr);
+		} else {
+#ifdef USE_DEB
+			MD5(mapped, size, (unsigned char *)&hptr);
+#else
+			;
+#endif
+		}
 		munmap(mapped, size);
 		digest = malloc((SHA256_LEN * 2) + 1);
 
 		// Convert to ASCII string
-		bytes2hex(digest, hptr, SHA256_LEN);
+		bytes2hex(digest, hptr, digest_length);
 	}
 	return digest;
 }
@@ -484,30 +513,24 @@ static int read_preliminary_header(int fd)
 }
 
 
-static Elf32_Ehdr *read_header32(int fd) MALLOCLIKE;
-static Elf32_Ehdr *read_header32(int fd)
+static Elf32_Ehdr *read_header32(int fd, Elf32_Ehdr *ptr)
 {
-	Elf32_Ehdr *ptr = malloc(sizeof(Elf32_Ehdr));
 	memcpy(ptr->e_ident, e_ident, EI_NIDENT);
 	ssize_t rc = safe_read(fd, (char *)ptr + EI_NIDENT,
 				sizeof(Elf32_Ehdr) - EI_NIDENT);
 	if (rc == (sizeof(Elf32_Ehdr) - EI_NIDENT))
 		return ptr;
-	free(ptr);
 	return NULL;
 }
 
 
-static Elf64_Ehdr *read_header64(int fd) MALLOCLIKE;
-static Elf64_Ehdr *read_header64(int fd)
+static Elf64_Ehdr *read_header64(int fd, Elf64_Ehdr *ptr)
 {
-	Elf64_Ehdr *ptr = malloc(sizeof(Elf64_Ehdr));
 	memcpy(ptr->e_ident, e_ident, EI_NIDENT);
 	ssize_t rc = safe_read(fd, (char *)ptr + EI_NIDENT,
 				sizeof(Elf64_Ehdr) - EI_NIDENT);
 	if (rc == (sizeof(Elf64_Ehdr) - EI_NIDENT))
 		return ptr;
-	free(ptr);
 	return NULL;
 }
 
@@ -544,8 +567,9 @@ uint32_t gather_elf(int fd, off_t size)
 	if (e_ident[EI_CLASS] == ELFCLASS32) {
 		unsigned i, type;
 		Elf32_Phdr *ph_tbl = NULL;
+		Elf32_Ehdr hdr_buf;
 
-		Elf32_Ehdr *hdr = read_header32(fd);
+		Elf32_Ehdr *hdr = read_header32(fd, &hdr_buf);
 		if (hdr == NULL) {
 			info |= HAS_ERROR;
 			goto rewind_out;
@@ -572,12 +596,10 @@ uint32_t gather_elf(int fd, off_t size)
 		if ((unsigned)hdr->e_phentsize != sizeof(Elf32_Phdr) ||
 		    (unsigned)hdr->e_phnum == 0) {
 			info |= HAS_ERROR;
-			free(hdr);
 			goto rewind_out;
 		}
 		if (sz > ((unsigned long)size - sizeof(Elf32_Ehdr))) {
 			info |= HAS_ERROR;
-			free(hdr);
 			goto rewind_out;
 		}
 		ph_tbl = malloc(sz);
@@ -696,12 +718,13 @@ err_out32:
 done32:
 		free(ph_tbl);
 done32_obj:
-		free(hdr);
+		; // fix an 'error label at end of compound statement'
 	} else if (e_ident[EI_CLASS] == ELFCLASS64) {
 		unsigned i, type;
 		Elf64_Phdr *ph_tbl;
+		Elf64_Ehdr hdr_buf;
 
-		Elf64_Ehdr *hdr = read_header64(fd);
+		Elf64_Ehdr *hdr = read_header64(fd, &hdr_buf);
 		if (hdr == NULL) {
 			info |= HAS_ERROR;
 			goto rewind_out;
@@ -728,12 +751,10 @@ done32_obj:
 		if ((unsigned)hdr->e_phentsize != sizeof(Elf64_Phdr) ||
 		    (unsigned)hdr->e_phnum == 0) {
 			info |= HAS_ERROR;
-			free(hdr);
 			goto rewind_out;
 		}
 		if (sz > ((unsigned long)size - sizeof(Elf64_Ehdr))) {
 			info |= HAS_ERROR;
-			free(hdr);
 			goto rewind_out;
 		}
 		ph_tbl = malloc(sz);
@@ -849,7 +870,7 @@ err_out64:
 done64:
 		free(ph_tbl);
 done64_obj:
-		free(hdr);
+		; // fix an 'error label at end of compound statement'
 	} else // Invalid ELF class
 		info |= HAS_ERROR;
 rewind_out:
